@@ -26,15 +26,18 @@ some set of rules.
 """
 
 import abc
+import functools
 import itertools
 import logging
+import operator
 import os
 import re
 
 from plainbox.abc import IJobQualifier
-from plainbox.impl.secure.rfc822 import FileTextSource
-from plainbox.impl.secure.rfc822 import Origin
-from plainbox.impl.secure.rfc822 import UnknownTextSource
+from plainbox.i18n import gettext as _
+from plainbox.impl.secure.origin import FileTextSource
+from plainbox.impl.secure.origin import Origin
+from plainbox.impl.secure.origin import UnknownTextSource
 
 
 _logger = logging.getLogger("plainbox.secure.qualifiers")
@@ -47,8 +50,15 @@ class SimpleQualifier(IJobQualifier):
     have share some code.
     """
 
-    def __init__(self, inclusive=True):
+    def __init__(self,  origin, inclusive=True):
+        if origin is not None and not isinstance(origin, Origin):
+            raise TypeError(_('argument {!a}, expected {}, got {}').format(
+                'origin', Origin, type(origin)))
+        if not isinstance(inclusive, bool):
+            raise TypeError(_('argument {!a}, expected {}, got {}').format(
+                'inclusive', bool, type(inclusive)))
         self._inclusive = inclusive
+        self._origin = origin
 
     @property
     def inclusive(self):
@@ -107,6 +117,16 @@ class SimpleQualifier(IJobQualifier):
         """
         return [self]
 
+    @property
+    def origin(self):
+        """
+        Origin of this qualifier
+
+        This property can be used to trace the origin of a qualifier back to
+        its definition point.
+        """
+        return self._origin
+
 
 class RegExpJobQualifier(SimpleQualifier):
     """
@@ -114,11 +134,11 @@ class RegExpJobQualifier(SimpleQualifier):
     expression
     """
 
-    def __init__(self, pattern, inclusive=True):
+    def __init__(self, pattern, origin, inclusive=True):
         """
         Initialize a new RegExpJobQualifier with the specified pattern.
         """
-        super().__init__(inclusive)
+        super().__init__(origin, inclusive)
         self._pattern = re.compile(pattern)
         self._pattern_text = pattern
 
@@ -148,10 +168,17 @@ class JobIdQualifier(SimpleQualifier):
     A JobQualifier that designates a single job with a particular id
     """
 
-    def __init__(self, id, inclusive=True):
-        super().__init__(inclusive)
+    def __init__(self, id, origin, inclusive=True):
+        super().__init__(origin, inclusive)
         self._id = id
 
+    @property
+    def id(self):
+        """
+        identifier to match
+        """
+        return self._id
+
     def get_simple_match(self, job):
         """
         Check if the given job matches this qualifier.
@@ -171,8 +198,8 @@ class NonLocalJobQualifier(SimpleQualifier):
     A JobQualifier that designates only non local jobs
     """
 
-    def __init__(self, inclusive=True):
-        super().__init__(inclusive)
+    def __init__(self, origin, inclusive=True):
+        super().__init__(origin, inclusive)
 
     def get_simple_match(self, job):
         """
@@ -205,6 +232,7 @@ class IMatcher(metaclass=abc.ABCMeta):
         """
 
 
+@functools.total_ordering
 class OperatorMatcher(IMatcher):
     """
     A matcher that applies a binary operator to the value
@@ -214,6 +242,25 @@ class OperatorMatcher(IMatcher):
         self._op = op
         self._value = value
 
+    @property
+    def op(self):
+        """
+        the operator to use
+
+        The operator is typically one of the functions from the ``operator``
+        module. For example. operator.eq corresponds to the == python operator.
+        """
+        return self._op
+
+    @property
+    def value(self):
+        """
+        The right-hand-side value to apply to the operator
+
+        The left-hand-side is the value that is passed to :meth:`match()`
+        """
+        return self._value
+
     def match(self, value):
         return self._op(self._value, value)
 
@@ -221,6 +268,22 @@ class OperatorMatcher(IMatcher):
         return "{0}({1!r}, {2!r})".format(
             self.__class__.__name__, self._op, self._value)
 
+    def __eq__(self, other):
+        if isinstance(other, OperatorMatcher):
+            return self.op == other.op and self.value == other.value
+        else:
+            return NotImplemented
+
+    def __lt__(self, other):
+        if isinstance(other, OperatorMatcher):
+            if self.op < other.op:
+                return True
+            if self.value < other.value:
+                return True
+            return False
+        else:
+            return NotImplemented
+
 
 class PatternMatcher(IMatcher):
     """
@@ -231,6 +294,10 @@ class PatternMatcher(IMatcher):
         self._pattern_text = pattern
         self._pattern = re.compile(pattern)
 
+    @property
+    def pattern_text(self):
+        return self._pattern_text
+
     def match(self, value):
         return self._pattern.match(value) is not None
 
@@ -238,13 +305,25 @@ class PatternMatcher(IMatcher):
         return "{0}({1!r})".format(
             self.__class__.__name__, self._pattern_text)
 
+    def __eq__(self, other):
+        if isinstance(other, PatternMatcher):
+            return self.pattern_text == other.pattern_text
+        else:
+            return NotImplemented
+
+    def __lt__(self, other):
+        if isinstance(other, PatternMatcher):
+            return self.pattern_text < other.pattern_text
+        else:
+            return NotImplemented
+
 
 class FieldQualifier(SimpleQualifier):
     """
     A SimpleQualifer that uses matchers to compare particular fields
     """
 
-    def __init__(self, field, matcher, inclusive=True):
+    def __init__(self, field, matcher, origin, inclusive=True):
         """
         Initialize a new FieldQualifier with the specified field, matcher and
         inclusive flag
@@ -256,10 +335,24 @@ class FieldQualifier(SimpleQualifier):
         :param inclusive:
             Inclusive selection flag (default: True)
         """
-        super().__init__(inclusive)
+        super().__init__(origin, inclusive)
         self._field = field
         self._matcher = matcher
 
+    @property
+    def field(self):
+        """
+        Name of the field to match
+        """
+        return self._field
+
+    @property
+    def matcher(self):
+        """
+        The IMatcher-implementing object to use to check for the match
+        """
+        return self._matcher
+
     def get_simple_match(self, job):
         """
         Check if the given job matches this qualifier.
@@ -317,6 +410,17 @@ class CompositeQualifier(IJobQualifier):
     def get_primitive_qualifiers(self):
         return get_flat_primitive_qualifier_list(self.qualifier_list)
 
+    @property
+    def origin(self):
+        raise NonPrimitiveQualifierOrigin
+
+
+class NonPrimitiveQualifierOrigin(Exception):
+    """
+    Exception raised when IJobQualifier.origin is meaningless as it is being
+    requested on a non-primitive qualifier such as the CompositeQualifier
+    """
+
 
 # NOTE: using CompositeQualifier seems strange but it's a tested proven
 # component so all we have to ensure is that we read the whitelist files
@@ -358,12 +462,13 @@ class WhiteList(CompositeQualifier):
                 else:
                     return maybe_partial_id_pattern
             qualifier_list = [
-                RegExpJobQualifier(transform_pattern(pattern), inclusive=True)
+                RegExpJobQualifier(
+                    transform_pattern(pattern), origin, inclusive=True)
                 for pattern in pattern_list]
         else:
             # Otherwise just use the patterns directly
             qualifier_list = [
-                RegExpJobQualifier(pattern, inclusive=True)
+                RegExpJobQualifier(pattern, origin, inclusive=True)
                 for pattern in pattern_list]
         super().__init__(qualifier_list)
 
@@ -590,10 +695,19 @@ def select_jobs(job_list, qualifier_list):
     # As a separate feature, we might return a list of qualifiers that never
     # matched anything. That may be helpful for debugging.
     included_list = []
+    id_to_index_map = {job.id: index for index, job in enumerate(job_list)}
     included_set = set()
     excluded_set = set()
     for qualifier in flat_qualifier_list:
-        for j_index, job in enumerate(job_list):
+        if (isinstance(qualifier, FieldQualifier)
+                and qualifier.field == 'id'
+                and isinstance(qualifier.matcher, OperatorMatcher)
+                and qualifier.matcher.op == operator.eq):
+            # optimize the super-common case where a qualifier refers to
+            # a specific job by using the id_to_index_map to instantly
+            # perform the requested operation on a single job
+            j_index = id_to_index_map(qualifier.matcher.value)
+            job = job_list[j_index]
             vote = qualifier.get_vote(job)
             if vote == IJobQualifier.VOTE_INCLUDE:
                 if j_index in included_set:
@@ -604,5 +718,17 @@ def select_jobs(job_list, qualifier_list):
                 excluded_set.add(j_index)
             elif vote == IJobQualifier.VOTE_IGNORE:
                 pass
+        else:
+            for j_index, job in enumerate(job_list):
+                vote = qualifier.get_vote(job)
+                if vote == IJobQualifier.VOTE_INCLUDE:
+                    if j_index in included_set:
+                        continue
+                    included_set.add(j_index)
+                    included_list.append(j_index)
+                elif vote == IJobQualifier.VOTE_EXCLUDE:
+                    excluded_set.add(j_index)
+                elif vote == IJobQualifier.VOTE_IGNORE:
+                    pass
     return [job_list[index] for index in included_list
             if index not in excluded_set]
